import os
from datetime import datetime, date
from flask import Flask, request, render_template, send_from_directory
import json
from json import JSONEncoder

from terka import bootstrap
from terka.domain import commands
from terka.service_layer import views, unit_of_work
from terka.utils import load_config

app = Flask(__name__)
app.config["JSONIFY_PRETTYPRINT_REGULAR"] = True

HOME_DIR = os.path.expanduser("~")
DB_URL = f"sqlite:////{HOME_DIR}/.terka/tasks.db"
STATIC_DIR = os.getenv("STATIC_DIR") or "static"
config = load_config(HOME_DIR)

bus = bootstrap.bootstrap(start_orm=True,
                          uow=unit_of_work.SqlAlchemyUnitOfWork(DB_URL),
                          config=config)


class EntityEncoder(JSONEncoder):

    def default(self, o):
        if isinstance(o, date):
            return o.strftime("%Y-%m-%d %H:%M:%s")
        else:
            return o.__dict__


@app.route("/", defaults={"path": ""})
@app.route("/<path:path>")
def catch_all(path):
    file_requested = os.path.join(app.root_path, STATIC_DIR, path)
    if not os.path.isfile(file_requested):
        path = "index.html"
    max_age = 0 if path == "index.html" else None
    return send_from_directory(STATIC_DIR, path, max_age=max_age)


# projects
@app.route("/api/v1/projects", methods=["GET"])
def projects():
    return _build_response(views.projects(bus.uow))


@app.route("/api/v1/projects/<project_id>", methods=["GET"])
def project(project_id):
    return _build_response(views.project(bus.uow, project_id))


@app.route("/api/v1/project", methods=["POST"])
def new_project():
    data = request.get_json(force=True)
    cmd = commands.CreateProject.from_kwargs(**data)
    result = bus.handle(cmd)
    return _build_response(result)


@app.route("/api/v1/projects/<project_id>", methods=["DELETE"])
def delete_project(project_id):
    cmd = commands.DeleteProject(project_id)
    result = bus.handle(cmd)
    return _build_response(result)


# tasks
@app.route("/api/v1/tasks", methods=["GET"])
def tasks():
    return _build_response(views.tasks(bus.uow))


@app.route("/api/v1/tasks/<task_id>", methods=["GET"])
def task(task_id):
    return _build_response(views.task(bus.uow, task_id))


@app.route("/api/v1/tasks/<task_id>/commentaries", methods=["GET"])
def task_commentaries(task_id):
    return _build_response(views.task_commentaries(bus.uow, task_id))


@app.route("/api/v1/task", methods=["POST"])
def new_task():
    data = request.get_json(force=True)
    cmd = commands.CreateTask.from_kwargs(**data)
    result = bus.handle(cmd)
    return _build_response(result)


@app.route("/api/v1/tasks/<task_id>", methods=["DELETE"])
def delete_task(task_id):
    cmd = commands.DeleteTask(task_id)
    result = bus.handle(cmd)
    return _build_response(result)


@app.route("/api/v1/tasks/<task_id>/comment", methods=["POST"])
def comment_task(task_id):
    data = request.get_json(force=True)
    data.update({"id": task_id})
    cmd = commands.CommentTask.from_kwargs(**data)
    result = bus.handle(cmd)
    return _build_response(result)


# workspaces
@app.route("/api/v1/workspaces", methods=["GET"])
def workspaces():
    return _build_response(views.workspaces(bus.uow))


@app.route("/api/v1/workspaces/<workspace_id>", methods=["GET"])
def workspace(workspace_id):
    return _build_response(views.workspace(bus.uow, workspace_id))


@app.route("/api/v1/workspaces/<workspace_id>/projects", methods=["GET"])
def workspace_projects(workspace_id):
    return _build_response(views.workspace_projects(bus.uow, workspace_id))


@app.route("/api/v1/workspace", methods=["POST"])
def new_workspace():
    data = request.get_json(force=True)
    cmd = commands.CreateWorkspace.from_kwargs(**data)
    result = bus.handle(cmd)
    return _build_response(result)


@app.route("/api/v1/workspaces/<workspace_id>", methods=["DELETE"])
def delete_workspace(workspace_id):
    cmd = commands.DeleteWorkspace(workspace_id)
    result = bus.handle(cmd)
    return _build_response(result)


# epics
@app.route("/api/v1/epics", methods=["GET"])
def epics():
    return _build_response(views.epics(bus.uow))


@app.route("/api/v1/epics/<epic_id>", methods=["GET"])
def epic(epic_id):
    return _build_response(views.epic(bus.uow, epic_id))


@app.route("/api/v1/epics/<epic_id>/tasks", methods=["GET"])
def epic_tasks(epic_id):
    return _build_response(views.epic_tasks(bus.uow, epic_id))


@app.route("/api/v1/epic", methods=["POST"])
def new_epic():
    data = request.get_json(force=True)
    cmd = commands.CreateEpic.from_kwargs(**data)
    result = bus.handle(cmd)
    return _build_response(result)


@app.route("/api/v1/epics/<epic_id>", methods=["DELETE"])
def delete_epic(epic_id):
    cmd = commands.DeleteEpic(epic_id)
    result = bus.handle(cmd)
    return _build_response(result)


#stories
@app.route("/api/v1/stories", methods=["GET"])
def stories():
    return _build_response(views.stories(bus.uow))


@app.route("/api/v1/stories/<story_id>", methods=["GET"])
def story(story_id):
    return _build_response(views.story(bus.uow, story_id))


@app.route("/api/v1/stories/<story_id>/tasks", methods=["GET"])
def story_tasks(story_id):
    return _build_response(views.story_tasks(bus.uow, story_id))


@app.route("/api/v1/story", methods=["POST"])
def new_story():
    data = request.get_json(force=True)
    cmd = commands.CreateStory.from_kwargs(**data)
    result = bus.handle(cmd)
    return _build_response(result)


@app.route("/api/v1/stories/<story_id>", methods=["DELETE"])
def delete_story(story_id):
    cmd = commands.DeleteStory(story_id)
    result = bus.handle(cmd)
    return _build_response(result)


# sprints
@app.route("/api/v1/sprints", methods=["GET"])
def sprints():
    return _build_response(views.sprints(bus.uow))


@app.route("/api/v1/sprints/<sprint_id>", methods=["GET"])
def sprint(sprint_id):
    return _build_response(views.sprint(bus.uow, sprint_id))


@app.route("/api/v1/sprints/<sprint_id>/tasks", methods=["GET"])
def sprint_tasks(sprint_id):
    return _build_response(views.sprint_tasks(bus.uow, sprint_id))


@app.route("/api/v1/sprint", methods=["POST"])
def new_sprint():
    data = request.get_json(force=True)
    cmd = commands.CreateSprint.from_kwargs(**data)
    result = bus.handle(cmd)
    return _build_response(result)


@app.route("/api/v1/sprints/<sprint_id>", methods=["DELETE"])
def delete_sprint(sprint_id):
    cmd = commands.DeleteSprint(sprint_id)
    result = bus.handle(cmd)
    return _build_response(result)


# users
@app.route("/api/v1/users", methods=["GET"])
def users():
    return _build_response(views.users(bus.uow))


@app.route("/api/v1/users/<user_id>", methods=["GET"])
def user(user_id):
    return _build_response(views.user(bus.uow, user_id))


@app.route("/api/v1/user", methods=["POST"])
def new_user():
    data = request.get_json(force=True)
    cmd = commands.CreateUser.from_kwargs(**data)
    result = bus.handle(cmd)
    return _build_response(result)


@app.route("/api/v1/users/<user_id>", methods=["DELETE"])
def delete_user(user_id):
    cmd = commands.DeleteUser(user_id)
    result = bus.handle(cmd)
    return _build_response(result)


def _build_response(msg="", status=200, mimetype="application/json"):
    """Helper method to build the response."""
    msg = json.dumps(msg, indent=4, cls=EntityEncoder)
    response = app.response_class(msg, status=status, mimetype=mimetype)
    response.headers["Access-Control-Allow-Origin"] = "*"
    return response


if __name__ == "__main__":
    app.run(debug=True, port=5000)



import pytest
from datetime import datetime, timedelta
import terka
from terka import bootstrap
from terka.domain import commands, entities, events


class TestTask:

    def test_create_simple_task(self, bus):
        cmd = commands.CreateTask(name="test")
        bus.handle(cmd)
        new_task = bus.uow.tasks.get_by_id(entities.task.Task, 1)
        expected_task = entities.task.Task(
            name="test",
            description=None,
            project=None,
            assignee=None,
            due_date=None,
            status=entities.task.TaskStatus.BACKLOG,
            priority=entities.task.TaskPriority.NORMAL)
        assert new_task == expected_task

    def test_updating_task_sets_modification_date(self, bus):
        cmd = commands.CreateTask(name="test")
        task_id = bus.handle(cmd)
        new_task = bus.uow.tasks.get_by_id(entities.task.Task, task_id)
        update_cmd = commands.UpdateTask(new_task.id, status="TODO")
        bus.handle(update_cmd)
        new_task = bus.uow.tasks.get_by_id(entities.task.Task, task_id)
        assert new_task.modification_date > new_task.creation_date
        assert new_task.status == entities.task.TaskStatus.TODO

    def test_completing_task_creates_status_task_event(self, bus):
        cmd = commands.CreateTask(name="test")
        task_id = bus.handle(cmd)
        complete_cmd = commands.CompleteTask(task_id)
        bus.handle(complete_cmd)
        new_task = bus.uow.tasks.get_by_id(entities.task.Task, task_id)
        task_event = bus.uow.tasks.get_by_conditions(
            entities.event_history.TaskEvent, {
                "task": task_id,
                "new_value": "DONE",
                "old_value": "BACKLOG"
            })
        assert task_event
        assert new_task.status == entities.task.TaskStatus.DONE

    def test_deleting_task_creates_status_task_event(self, bus):
        cmd = commands.CreateTask(name="test")
        task_id = bus.handle(cmd)
        complete_cmd = commands.DeleteTask(task_id)
        bus.handle(complete_cmd)
        new_task = bus.uow.tasks.get_by_id(entities.task.Task, task_id)
        task_event = bus.uow.tasks.get_by_conditions(
            entities.event_history.TaskEvent, {
                "task": task_id,
                "new_value": "DELETED",
                "old_value": "BACKLOG"
            })
        assert task_event
        assert new_task.status == entities.task.TaskStatus.DELETED

    def test_creating_task_with_tag_creates_tag(self, bus):
        cmd = commands.CreateTask(name="test")
        task_id = bus.handle(cmd, context={"tags": "new_tag"})
        new_task_tag = bus.uow.tasks.get_by_conditions(
            entities.tag.TaskTag, {"task": task_id})
        new_base_tag = bus.uow.tasks.get_by_conditions(
            entities.tag.BaseTag, {"text": "new_tag"})
        assert new_task_tag
        assert new_base_tag

    def test_creating_task_with_collaborator_creates_user(self, bus):
        cmd = commands.CreateTask(name="test")
        task_id = bus.handle(cmd,
                             context={"collaborators": "new_collaborator"})
        [new_task_collaborator] = bus.uow.tasks.get_by_conditions(
            entities.collaborator.TaskCollaborator, {"task": task_id})
        new_user = bus.uow.tasks.get_by_id(entities.user.User,
                                                   new_task_collaborator.id)
        assert new_task_collaborator
        assert new_user

    def test_creating_task_with_comment_creates_comment(self, bus):
        cmd = commands.CreateTask(name="test")
        task_id = bus.handle(cmd, context={"comment": "new_commentary"})
        new_task_comment = bus.uow.tasks.get_by_conditions(
            entities.commentary.TaskCommentary, {"task": task_id})
        assert new_task_comment

    def test_commenting_task_with_empty_comment_is_ignored(self, bus):
        cmd = commands.CreateTask(name="test")
        task_id = bus.handle(cmd, context={"comment": " "})
        new_task_comment = bus.uow.tasks.get_by_conditions(
            entities.commentary.TaskCommentary, {"task": task_id})
        assert not new_task_comment

    def test_tagging_task_with_empty_tag_is_ignored(self, bus):
        cmd = commands.CreateTask(name="test")
        task_id = bus.handle(cmd, context={"tag": " "})
        new_task_tag = bus.uow.tasks.get_by_conditions(
            entities.tag.TaskTag, {"task": task_id})
        assert not new_task_tag

    @pytest.mark.parametrize("entity", ["epic", "sprint", "story"])
    def test_tagging_task_with_service_tag_is_ignored(self, bus, entity):
        cmd = commands.CreateTask(name="test")
        task_id = bus.handle(cmd, context={"tag": f"{entity}: 1"})
        new_task_tag = bus.uow.tasks.get_by_conditions(
            entities.tag.TaskTag, {"task": task_id})
        assert not new_task_tag


class TestSprint:

    @pytest.fixture(scope="class")
    def tasks(self, bus):
        create_task_1 = commands.CreateTask(name="task_1")
        create_task_2 = commands.CreateTask(name="task_2")
        task_1 = bus.handle(create_task_1)
        task_2 = bus.handle(create_task_2)
        return task_1, task_2

    @pytest.fixture
    def new_sprint(self, bus):
        today = datetime.now()
        next_monday = (today + timedelta(days=(7 - today.weekday())))
        next_sunday = (today + timedelta(days=(13 - today.weekday())))
        cmd = commands.CreateSprint(start_date=next_monday,
                                     end_date=next_sunday)
        sprint_id = bus.handle(cmd)
        return sprint_id

    @pytest.fixture
    def sprint_with_tasks(self, bus, new_sprint, tasks):
        task_1, task_2 = tasks
        add_task_1 = commands.AddTask(id=task_1, sprint=new_sprint)
        add_task_2 = commands.AddTask(id=task_2, sprint=new_sprint)
        bus.handle(add_task_1)
        bus.handle(add_task_2)

    def test_new_sprint_created(self, new_sprint):
        assert new_sprint

    def test_referencing_non_existing_sprints_raises_entity_not_found_exception(
            self, bus):
        with pytest.raises(terka.service_layer.exceptions.EntityNotFound):
            bus.handle(commands.StartSprint(9999))

    def test_cannot_create_sprint_with_end_date_in_past(self, bus):
        today = datetime.now()
        start_date = (today - timedelta(days=(7 - today.weekday())))
        end_date = (today - timedelta(days=(1 - today.weekday())))
        cmd = commands.CreateSprint(start_date=start_date, end_date=end_date)
        # TODO: check for a specific message
        with pytest.raises(ValueError):
            bus.handle(cmd)

    def test_adding_task_to_sprint(self, bus, new_sprint, sprint_with_tasks):
        sprint_tasks = bus.uow.tasks.get_by_conditions(
            entities.sprint.SprintTask, {"sprint": new_sprint})
        assert len(sprint_tasks) == 2
        for task in sprint_tasks:
            assert task.story_points == 0

    def test_starting_sprint_changes_status_due_date(self, bus, new_sprint,
                                                     sprint_with_tasks):
        cmd = commands.StartSprint(new_sprint)
        bus.handle(cmd)
        sprint = bus.uow.tasks.get_by_id(entities.sprint.Sprint,
                                                 new_sprint)
        assert sprint.status == entities.sprint.SprintStatus.ACTIVE
        sprint_tasks = bus.uow.tasks.get_by_conditions(
            entities.sprint.SprintTask, {"sprint": new_sprint})
        for sprint_task in sprint_tasks:
            task = bus.uow.tasks.get_by_id(entities.task.Task,
                                                   sprint_task.task)
            assert task.status == entities.task.TaskStatus.TODO
            assert task.due_date == sprint.end_date

    def test_cannot_start_active_sprint(self, bus, new_sprint):
        cmd = commands.StartSprint(new_sprint)
        bus.handle(cmd)
        cmd = commands.StartSprint(new_sprint)
        with pytest.raises(terka.service_layer.exceptions.TerkaSprintActive):
            bus.handle(cmd)

    def test_cannot_start_completed_sprint(self, bus, new_sprint):
        cmd = commands.StartSprint(new_sprint)
        bus.handle(cmd)
        cmd = commands.CompleteSprint(new_sprint)
        bus.handle(cmd)
        cmd = commands.StartSprint(new_sprint)
        with pytest.raises(
                terka.service_layer.exceptions.TerkaSprintCompleted):
            bus.handle(cmd)

    def test_completing_sprint_changes_status_due_date(self, bus, new_sprint,
                                                       sprint_with_tasks):
        cmd = commands.CompleteSprint(new_sprint)
        bus.handle(cmd)
        sprint = bus.uow.tasks.get_by_id(entities.sprint.Sprint,
                                                 new_sprint)
        assert sprint.status == entities.sprint.SprintStatus.COMPLETED
        sprint_tasks = bus.uow.tasks.get_by_conditions(
            entities.sprint.SprintTask, {"sprint": new_sprint})
        for sprint_task in sprint_tasks:
            task = bus.uow.tasks.get_by_id(entities.task.Task,
                                                   sprint_task.task)
            assert task.status == entities.task.TaskStatus.BACKLOG
            assert not task.due_date

    def test_completing_sprint_with_in_progress_review_tasks_doesnt_change_their_status(
            self, bus, new_sprint, sprint_with_tasks):
        cmd = commands.StartSprint(new_sprint)
        bus.handle(cmd)
        sprint = bus.uow.tasks.get_by_id(entities.sprint.Sprint,
                                                 new_sprint)
        update_task = commands.UpdateTask(id=1, status="REVIEW")
        bus.handle(update_task)
        sprint_tasks = bus.uow.tasks.get_by_conditions(
            entities.sprint.SprintTask, {"sprint": new_sprint})
        for sprint_task in sprint_tasks:
            task = bus.uow.tasks.get_by_id(entities.task.Task,
                                                   sprint_task.task)
            if task.id == 1:
                assert task.status == entities.task.TaskStatus.REVIEW

    def test_cannot_add_ask_to_completed_sprint(self, bus, new_sprint,
                                                sprint_with_tasks):
        cmd = commands.CompleteSprint(new_sprint)
        bus.handle(cmd)
        create_task_3 = commands.CreateTask(name="task_3")
        task_3 = bus.handle(create_task_3)
        cmd = commands.AddTask(id=task_3, sprint=new_sprint)
        with pytest.raises(
                terka.service_layer.exceptions.TerkaSprintCompleted):
            bus.handle(cmd)

    def test_deleting_task_from_sprint(self, bus, new_sprint,
                                       sprint_with_tasks):
        cmd = commands.StartSprint(new_sprint)
        bus.handle(cmd)
        sprint = bus.uow.tasks.get_by_id(entities.sprint.Sprint,
                                                 new_sprint)
        delete_task = commands.DeleteTask(id=sprint.tasks[0].task,
                                           sprint=new_sprint)
        bus.handle(delete_task)
        sprint = bus.uow.tasks.get_by_id(entities.sprint.Sprint,
                                                 new_sprint)
        assert len(sprint.tasks) == 1

    def test_adding_story_to_sprint_adds_non_completed_tasks(
            self, bus, new_sprint, tasks):
        task_1, task_2 = tasks
        create_story = commands.CreateStory(name="story_1")
        story_id = bus.handle(create_story)
        add_task_1 = commands.AddTask(id=task_1, story=story_id)
        add_task_2 = commands.AddTask(id=task_2, story=story_id)
        bus.handle(add_task_1)
        bus.handle(add_task_2)
        add_story_to_sprint = commands.AddStory(id=story_id,
                                                 sprint=new_sprint)
        bus.handle(add_story_to_sprint)
        sprint = bus.uow.tasks.get_by_id(entities.sprint.Sprint,
                                                 new_sprint)
        assert len(sprint.tasks) == 2

    def test_adding_epic_to_sprint_adds_non_completed_tasks(
            self, bus, new_sprint, tasks):
        task_1, task_2 = tasks
        create_epic = commands.CreateEpic(name="epic_1")
        epic_id = bus.handle(create_epic)
        add_task_1 = commands.AddTask(id=task_1, epic=epic_id)
        add_task_2 = commands.AddTask(id=task_2, epic=epic_id)
        bus.handle(add_task_1)
        bus.handle(add_task_2)
        add_epic_to_sprint = commands.AddEpic(id=epic_id, sprint=new_sprint)
        bus.handle(add_epic_to_sprint)
        sprint = bus.uow.tasks.get_by_id(entities.sprint.Sprint,
                                                 new_sprint)
        assert len(sprint.tasks) == 2

    def test_adding_story_to_sprint_adds_only_non_completed_tasks(
            self, bus, new_sprint, tasks):
        task_1, task_2 = tasks
        create_story = commands.CreateStory(name="story_1")
        story_id = bus.handle(create_story)
        add_task_1 = commands.AddTask(id=task_1, story=story_id)
        add_task_2 = commands.AddTask(id=task_2, story=story_id)
        bus.handle(add_task_1)
        bus.handle(add_task_2)
        complete_task = commands.CompleteTask(task_1)
        bus.handle(complete_task)
        add_story_to_sprint = commands.AddStory(id=story_id,
                                                 sprint=new_sprint)
        bus.handle(add_story_to_sprint)
        sprint = bus.uow.tasks.get_by_id(entities.sprint.Sprint,
                                                 new_sprint)
        assert len(sprint.tasks) == 1

    def test_adding_epic_to_sprint_adds_only_non_completed_tasks(
            self, bus, new_sprint, tasks):
        task_1, task_2 = tasks
        create_epic = commands.CreateEpic(name="epic_1")
        epic_id = bus.handle(create_epic)
        add_task_1 = commands.AddTask(id=task_1, epic=epic_id)
        add_task_2 = commands.AddTask(id=task_2, epic=epic_id)
        bus.handle(add_task_1)
        bus.handle(add_task_2)
        complete_task = commands.CompleteTask(task_1)
        bus.handle(complete_task)
        add_epic_to_sprint = commands.AddEpic(id=epic_id, sprint=new_sprint)
        bus.handle(add_epic_to_sprint)
        sprint = bus.uow.tasks.get_by_id(entities.sprint.Sprint,
                                                 new_sprint)
        assert len(sprint.tasks) == 1



from __future__ import annotations
from copy import copy
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
from sqlalchemy import text

from terka.domain import entities


def n_days_ago(n: int) -> str:
    return (datetime.today() - timedelta(n)).strftime("%Y-%m-%d")


def projects(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.project.Project)]


def projects_id_to_name_mapping(uow) -> dict[int, str]:
    with copy(uow) as uow:
        return {t.id: t.name for t in uow.tasks.list(entities.project.Project)}


def project(uow, project_id: int):
    with uow:
        if not (project := uow.tasks.get_by_id(entities.project.Project,
                                               project_id)):
            return {}
        tasks = [task.to_dict() for task in project.tasks]
        result = project.to_dict()
        result["tasks"] = tasks
        return result


def tasks(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.task.Task)]


def task(uow, task_id: int):
    with uow:
        if not (task := uow.tasks.get_by_id(entities.task.Task, task_id)):
            return {}
        result = task.to_dict()
        if commentaries := task.commentaries:
            commentaries = [comment.to_dict() for comment in commentaries]
        result["commentaries"] = commentaries
        return result


def task_commentaries(uow, task_id: int) -> dict:
    with uow:
        if not (task := uow.tasks.get_by_id(entities.task.Task, task_id)):
            return {}
        if commentaries := task.commentaries:
            return [comment.to_dict() for comment in commentaries]
        return {}


def workspaces(uow) -> list[dict]:
    with uow:
        return [
            t.to_dict() for t in uow.tasks.list(entities.workspace.Workspace)
        ]


def workspaces_id_to_name_mapping(uow) -> dict[int, str]:
    with copy(uow) as uow:
        return {
            t.id: t.name
            for t in uow.tasks.list(entities.workspace.Workspace)
        }


def workspace(uow, workspace_id: int):
    with uow:
        if not (workspace := uow.tasks.get_by_id(entities.workspace.Workspace,
                                                 workspace_id)):
            return {}
        if projects := workspace.projects:
            projects = [project.to_dict() for project in workspace.projects]
            result = workspace.to_dict()
        result["projects"] = projects
        return result


def workspace_projects(uow, workspace_id: int):
    with uow:
        if not (workspace := uow.tasks.get_by_id(entities.workspace.Workspace,
                                                 workspace_id)):
            return {}
        if projects := workspace.projects:
            return [project.to_dict() for project in workspace.projects]
        return {}


def epics(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.epic.Epic)]


def epic(uow, epic_id: int):
    with uow:
        if not (epic := uow.tasks.get_by_id(entities.epic.Epic, epic_id)):
            return {}
        if tasks := epic.tasks:
            tasks = [task.tasks.to_dict() for task in epic.tasks]
            result = epic.to_dict()
        result["tasks"] = tasks
        return result


def epic_tasks(uow, epic_id: int):
    with uow:
        if not (epic := uow.tasks.get_by_id(entities.epic.Epic, epic_id)):
            return {}
        if tasks := epic.tasks:
            return [task.tasks.to_dict() for task in epic.tasks]
        return {}


def stories(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.story.Story)]


def story(uow, story_id: int):
    with uow:
        if not (story := uow.tasks.get_by_id(entities.story.Story, story_id)):
            return {}
        if tasks := story.tasks:
            tasks = [task.tasks.to_dict() for task in story.tasks]
            result = story.to_dict()
        result["tasks"] = tasks
        return result


def story_tasks(uow, story_id: int):
    with uow:
        if not (story := uow.tasks.get_by_id(entities.story.Story, story_id)):
            return {}
        if tasks := story.tasks:
            return [task.tasks.to_dict() for task in story.tasks]
        return {}


def sprints(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.sprint.Sprint)]


def sprint(uow, sprint_id: int):
    with uow:
        if not (sprint := uow.tasks.get_by_id(entities.sprint.Sprint,
                                              sprint_id)):
            return {}
        if tasks := sprint.tasks:
            tasks = [task.tasks.to_dict() for task in sprint.tasks]
            result = sprint.to_dict()
        result["tasks"] = tasks
        return result


def sprint_tasks(uow, sprint_id: int):
    with uow:
        if not (sprint := uow.tasks.get_by_id(entities.sprint.Sprint,
                                              sprint_id)):
            return {}
        if tasks := sprint.tasks:
            return [task.tasks.to_dict() for task in sprint.tasks]
        return {}


def users(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.user.User)]


def users_id_to_name_mapping(uow) -> dict:
    with copy(uow) as uow:
        return {t.id: t.name for t in uow.tasks.list(entities.user.User)}


def user(uow, user_id: int):
    with uow:
        if not (user := uow.tasks.get_by_id(entities.user.User, user_id)):
            return {}
        return user.to_dict()


def time_spent(session,
               tasks: list[entities.task.Task] | None,
               start_date: str = n_days_ago(7),
               end_date: str = n_days_ago(-1),
               excluded_tasks_only: bool = False) -> list[dict[str, str]]:
    tasks = ",".join([str(task.id) for task in tasks])
    query = f"""
    SELECT
        STRFTIME("%Y-%m-%d", creation_date) AS date,
        SUM(time_spent_minutes) AS time_spent_hours
    FROM time_tracker_entries
    WHERE
        creation_date >= "{start_date}"
        AND creation_date <= "{end_date}"
    """
    if excluded_tasks_only:
        query += f" AND task NOT IN ({tasks})"
    else:
        query += f" AND task IN ({tasks})"
    query += "GROUP BY 1"
    results = session.execute(text(query))
    return [dict(r) for r in results]


def sprint_task_ids(session,
                    sprint_id: int | None = None) -> list[dict[int, int]]:
    results = session.execute(
        """
    SELECT
        task, sprint
    FROM sprint_tasks
    WHERE
        sprint = :sprint_id
    """, dict(sprint_id=sprint_id))
    return [dict(r) for r in results]


def external_connectors_asana_projects(session) -> list[dict[int, str]]:
    results = session.execute("""
    SELECT
        id, asana_project_id, sync_date
    FROM 'external_connectors.asana.projects'
    """)
    return [dict(r) for r in results]

def external_connectors_asana_users(session) -> dict[int, str]:
    results = session.execute("""
    SELECT
        id, asana_user_id
    FROM 'external_connectors.asana.users'
    """)
    return {r.id: r.asana_user_id for r in results}

def external_connectors_asana_tasks(
        session,
        project_id: str) -> dict[int, dict[str, str | datetime | None]]:
    results = session.execute(
        """
    SELECT
        id, asana_task_id, sync_date
    FROM 'external_connectors.asana.tasks'
    WHERE project = :project_id
    """, dict(project_id=project_id))
    return {
        r.id: {
            "asana_task_id":
            r.asana_task_id,
            "sync_date":
            datetime.strptime(r.sync_date, "%Y-%m-%d %H:%M:%S.%f")
            if r.sync_date else None
        }
        for r in results
    }


def status_changes(session, sprint_id: int):
    query = f"""
        SELECT DISTINCT T.id AS task,
        T.status AS current_status,
        STRFTIME("%Y-%m-%d", E.date) AS date,
        FIRST_VALUE(E.old_value) OVER (PARTITION BY T.id, STRFTIME("%Y-%m-%d", E.date) ORDER by E.date RANGE BETWEEN UNBOUNDED PRECEDING AND
                UNBOUNDED FOLLOWING) AS first_status_for_date,
        LAST_VALUE(E.new_value) OVER (PARTITION BY T.id, STRFTIME("%Y-%m-%d", E.date) ORDER by E.date RANGE BETWEEN UNBOUNDED PRECEDING AND
                UNBOUNDED FOLLOWING) AS last_status_for_date
        FROM tasks AS T
        LEFT JOIN (SELECT * FROM task_events WHERE type = "STATUS") AS E
            ON T.id = E.task
        WHERE T.id IN (SELECT task FROM sprint_tasks WHERE sprint = {sprint_id})

        """
    return pd.read_sql_query(query, session.get_bind()).replace({np.nan: None})



import pathlib
from itertools import chain
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__)
README = (HERE.parent / "README.md").read_text()
EXTRAS_REQUIRE = {"asana": ["asana==5.0.0"]}

EXTRAS_REQUIRE['all'] = list(set(chain(*EXTRAS_REQUIRE.values())))

setup(
    name="terka",
    version="2.7.0",
    description="CLI utility for creating and managing tasks in a terminal",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Andrei Markin",
    author_email="andrey.markin.ppc@gmail.com",
    license="Apache 2.0",
    url="https://github.com/AndreyMarkinPPC/terka",
    packages=find_packages(),
    include_package_data=True,
    package_data={"": ["service_layer/*.css"]},
    install_requires=[
        "sqlalchemy==1.4.0",
        "pyaml",
        "rich",
        "textual==0.41.0",
        "plotext==5.2.8",
        "textual-plotext==0.2.1",
        "pandas",
        "matplotlib",
    ],
    setup_requires=["pytest-runner"],
    tests_requires=["pytest"],
    extras_require=EXTRAS_REQUIRE,
    entry_points={"console_scripts": ["terka=terka.entrypoints.cli:main"]},
)



from __future__ import annotations
from collections import abc
from dataclasses import dataclass
from datetime import datetime
import inspect
import rich
from rich.console import Console
from rich.table import Table
from textual.app import App, ComposeResult
from textual import widgets

from terka.service_layer import exceptions, formatter, ui, views


@dataclass
class PrintOptions:
    show_tasks: bool = True
    show_history: bool = False
    show_commentaries: bool = False
    show_completed: bool = False
    show_epics: bool = True
    show_stories: bool = True
    show_notes: bool = True
    show_viz: bool = False
    sort: str = "id"
    columns: str = ""
    expand_table: bool = True

    @classmethod
    def from_kwargs(cls, **kwargs: dict) -> "PrintOptions":
        return cls(**{k: kwargs[k] for k in kwargs if k in cls.__match_args__})


class Printer:

    def __init__(self, uow):
        self.console = ConsolePrinter(uow)
        self.tui = TextualPrinter(uow)


class TextualPrinter:

    def __init__(self, uow) -> None:
        self.uow = uow

    def show_note(self, note):

        class NoteMarkdownViewer(App):

            BINDINGS = [("q", "quit", "Quit")]

            def compose(self) -> ComposeResult:
                yield widgets.MarkdownViewer(str(note),
                                             show_table_of_contents=True)

        app = NoteMarkdownViewer()
        app.run()

    def print_project(self, project, bus):
        app = ui.TerkaProject(project, bus)
        app.run()
        if app.return_code == 4:
            raise exceptions.TerkaRefreshException


    def print_sprint(self, sprint, bus):
        app = ui.TerkaSprint(sprint, bus)
        app.run()
        if app.return_code == 4:
            raise exceptions.TerkaRefreshException

class ConsolePrinter:

    def __init__(self, uow, box=rich.box.SIMPLE, config=None) -> None:
        self.uow = uow
        self.box = box
        self.config = config
        self.console = Console()

    def print_new_object(self, obj, mappings: dict | None = None):
        table = Table(box=self.box)
        if not mappings:
            mappings = {
                "project_mapping": views.projects_id_to_name_mapping(self.uow),
                "user_mapping": views.users_id_to_name_mapping(self.uow),
                "workspace_mapping": views.workspaces_id_to_name_mapping(self.uow)

            }
        attributes = self._get_attributes(obj, mappings)
        for column, value in attributes:
            if value:
                table.add_column(column)
        table.add_row(*list(zip(*attributes))[1])
        if table.row_count:
            self.console.print(f"Added new {obj.__class__.__name__}")
            self.console.print(table)

    def print_tag(self, entities):
        table = Table(box=self.box)
        for column in ("id", "text"):
            table.add_column(column)
        seen_tags = set()
        for entity in entities:
            text = entity.text
            if text and text not in seen_tags and not text.startswith(
                ("sprint:", "epic:", "story:")):
                table.add_row(f"[red]{entity.id}[/red]", entity.text)
                seen_tags.add(entity.text)
        if table.row_count:
            self.console.print(table)

    def print_user(self, entities):
        table = Table(box=self.box)
        for column in ("id", "name"):
            table.add_column(column)
        seen_users = set()
        for entity in entities:
            name = entity.name
            if name and name not in seen_users:
                table.add_row(f"[red]{entity.id}[/red]", name)
                seen_users.add(name)
        if table.row_count:
            self.console.print(table)

    def print_composite(
            self,
            entities,
            print_options,
            composite_type,
            kwargs=None):
        if not entities:
            self.console.print(f"[red]No {composite_type} found[/red]")
            exit()
        project_mapping = views.projects_id_to_name_mapping(self.uow)
        table = Table(box=self.box,
                      title=composite_type.upper(),
                      expand=print_options.expand_table)
        non_active_entities = Table(box=self.box,
                                    title=f"INACTIVE {composite_type.upper()}",
                                    expand=print_options.expand_table)
        default_columns = ("id", "name", "description", "status", "project",
                           "tasks")
        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = default_columns
        for column in printable_columns:
            table.add_column(column, style="bold")
        for column in printable_columns:
            non_active_entities.add_column(column, style="bold")
        for i, entity in enumerate(entities):
            project = project_mapping.get(entity.project)
            printable_row = {
                "id": f"{entity.id}",
                "name": str(entity.name),
                "description": entity.description,
                "status": entity.status.name,
                "project": project,
                "tasks": str(len(entity.open_tasks))
            }
            printable_elements = [
                value for key, value in printable_row.items()
                if key in printable_columns
            ]
            if (entity.status.name == "ACTIVE"
                    or len(entity.completed_tasks) < len(entity.tasks)):
                table.add_row(*printable_elements)
            else:
                non_active_entities.add_row(*printable_elements)
        if table.row_count:
            self.console.print(table)
        if non_active_entities.row_count and print_options.show_completed:
            self.console.print(non_active_entities)
        if viz := print_options.show_viz:
            if "time" in viz:
                time_entries = entity.daily_time_entries_hours()
                self._print_time_utilization(time_entries)

    def print_project(self, entities, print_options, kwargs=None):
        if not entities:
            self.console.print("[red]No projects found[/red]")
            exit()
        table = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                      expand=print_options.expand_table)
        non_active_projects = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                                    expand=print_options.expand_table)
        default_columns = ("id", "name", "description", "status", "open_tasks",
                           "overdue", "stale", "backlog", "todo",
                           "in_progress", "review", "done", "median_task_age",
                           "time_spent")

        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = default_columns
        for column in printable_columns:
            table.add_column(column)
        for column in printable_columns:
            if column in ("id", "name", "description", "status", "open_tasks"):
                non_active_projects.add_column(column)
        try:
            reverse = True
            sorting_field = getattr(entities[0], print_options.sort)
            if isinstance(sorting_field, abc.MutableSequence):
                sort_fn = lambda x: len(getattr(x, print_options.sort))
            else:
                sort_fn = lambda x: getattr(x, print_options.sort)
        except AttributeError:
            sort_fn = lambda x: "id"
        if print_options.sort == "id":
            reverse = False
        entities.sort(key=sort_fn, reverse=reverse)
        for entity in entities:
            if len(open_tasks :=
                   entity.open_tasks) > 0 and entity.status.name == "ACTIVE":

                if overdue_tasks := entity.overdue_tasks:
                    entity_id = f"[red]{entity.id}[/red]"
                else:
                    entity_id = f"[green]{entity.id}[/green]"
                printable_row = {
                    "id":
                    f"{entity.id}",
                    "name":
                    str(entity.name),
                    "description":
                    entity.description,
                    "status":
                    entity.status.name,
                    "open_tasks":
                    str(len(open_tasks)),
                    "overdue":
                    str(len(overdue_tasks)),
                    "stale":
                    str(len(entity.stale_tasks)),
                    "backlog":
                    str(len(entity.backlog)),
                    "todo":
                    str(len(entity.todo)),
                    "in_progress":
                    str(len(entity.in_progress)),
                    "review":
                    str(len(entity.review)),
                    "done":
                    str(len(entity.done)),
                    "median_task_age":
                    str(entity.median_task_age),
                    "time_spent":
                    formatter.Formatter.format_time_spent(
                        entity.total_time_spent),
                }
                printable_elements = [
                    value for key, value in printable_row.items()
                    if key in printable_columns
                ]
                table.add_row(*printable_elements)
            else:
                printable_row = {
                    "id": f"{entity.id}",
                    "name": str(entity.name),
                    "description": entity.description,
                    "status": entity.status.name,
                    "open_tasks": str(len(open_tasks)),
                }
                printable_elements = [
                    value for key, value in printable_row.items()
                    if key in printable_columns
                ]
                non_active_projects.add_row(*printable_elements)

        if table.row_count:
            self.console.print(table)
        if non_active_projects.row_count and print_options.show_completed:
            self.console.print("[green]Inactive projects[/green]")
            self.console.print(non_active_projects)

    def print_task(self, entities, print_options):
        table = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                      expand=print_options.expand_table)
        columns = ("id", "name", "description", "status", "priority",
                   "project", "due_date", "tags", "collaborators",
                   "time_spent")
        completed_tasks_default_columns = ("id", "name", "description",
                                           "status", "priority", "project",
                                           "completed_date", "tags",
                                           "collaborators", "time_spent")
        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = columns

        for column in printable_columns:
            table.add_column(column)

        for task in sorted(entities, key=lambda x: x.id, reverse=True):
            if tags := task.tags:
                tags_text = ",".join([tag.base_tag.text for tag in list(tags)])
            else:
                tags_text = ""
            if collaborators := task.collaborators:
                collaborators_texts = sorted([
                    collaborator.users.name
                    for collaborator in list(collaborators)
                    if collaborator.users
                ])
                collaborator_string = ",".join(collaborators_texts)
            else:
                collaborator_string = ""
            if task.is_overdue:
                task_id = f"[red]{task.id}[/red]"
            elif task.is_stale:
                task_id = f"[yellow]{task.id}[/yellow]"
            else:
                task_id = str(task.id)
            printable_row = {
                "id":
                task_id,
                "name":
                task.name,
                "description":
                task.description,
                "status":
                task.status.name,
                "priority":
                task.priority.name,
                "project":
                str(task.project),
                "due_date":
                str(task.due_date),
                "tags":
                tags_text,
                "collaborators":
                collaborator_string,
                "time_spent":
                formatter.Formatter.format_time_spent(task.total_time_spent),
            }
            printable_elements = [
                value for key, value in printable_row.items()
                if key in printable_columns
            ]
            table.add_row(*printable_elements)
        self.console.print(table)

    def print_sprint(self, entities, print_options, kwargs=None):
        if not entities:
            self.console.print("[red]No sprints found[/red]")
            exit()
        table = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                      expand=print_options.expand_table)
        all_sprints = Table(box=rich.box.SQUARE_DOUBLE_HEAD,
                            expand=print_options.expand_table)
        columns = ("id", "start_date", "end_date", "goal", "status",
                   "open tasks", "pct_completed", "velocity", "collaborators",
                   "time_spent", "utilization")
        if print_options.columns:
            printable_columns = print_options.columns.split(",")
        else:
            printable_columns = columns

        for column in printable_columns:
            table.add_column(column)
            all_sprints.add_column(column)
        for entity in entities:
            printable_row = {
                "id":
                str(entity.id),
                "start_date":
                str(entity.start_date),
                "end_date":
                str(entity.end_date),
                "goal":
                entity.goal,
                "status":
                entity.status.name,
                "open tasks":
                f"{len(entity.open_tasks)} ({len(entity.tasks)})",
                "pct_completed":
                f"{round(entity.pct_completed * 100, 2)}%",
                "velocity":
                str(round(entity.velocity, 2)),
                "collaborators":
                entity.collaborators_as_string,
                "time_spent":
                str(
                    formatter.Formatter.format_time_spent(
                        entity.total_time_spent)),
                "utilization":
                f"{round(entity.utilization * 100)}%"
            }
            printable_elements = [
                value for key, value in printable_row.items()
                if key in printable_columns
            ]
            if entity.status.name in ("ACTIVE", "PLANNED"):
                table.add_row(*printable_elements)
            all_sprints.add_row(*printable_elements)

        if print_options.show_completed:
            self.console.print(all_sprints)
        elif table.row_count:
            self.console.print(table)

    def _get_attributes(
            self, obj, mappings: dict[dict[int,
                                           str]]) -> list[tuple[str, str]]:
        attributes = []
        for name, value in inspect.getmembers(obj):
            if not name.startswith("_") and not inspect.ismethod(value):
                if not value:
                    continue
                elif name in ("created_by", "assignee"):
                    user = mappings.get("user_mapping").get(value)
                    attributes.append((name, user))
                elif name == "project":
                    project = mappings.get("project_mapping").get(value)
                    attributes.append((name, project))
                elif name == "workspace":
                    workspace = mappings.get("workspace_mapping",
                                             {}).get(value)
                    attributes.append((name, workspace))
                elif hasattr(value, "name"):
                    attributes.append((name, value.name))
                elif isinstance(value, datetime):
                    attributes.append((name, value.strftime("%Y-%m-%d %H:%M")))
                elif isinstance(value, set):
                    values = value.pop()
                    attributes.append((name, str(values)))
                else:
                    attributes.append((name, str(value)))
        return attributes

    def _print_time_utilization(self, time_entries) -> None:
        plt.date_form('Y-m-d')
        plt.plot_size(100, 15)
        plt.title(
            f"Time tracker - {format_hour_minute(sum(time_entries.values()))} spent"
        )
        plt.bar(time_entries.keys(), time_entries.values())
        plt.show()



from datetime import datetime

from .entity import Entity


class User(Entity):

    def __init__(self, name: str, **kwargs):
        self.name = name

    def __repr__(self):
        return f"<User {self.name}>"

    def __eq__(self, other):
       if not isinstance(other, User):
           return False
       return other.name == self.name
